import json
import os
import time
import traceback
import functools
import sys
import asyncio
import signal
import random
import concurrent.futures
import queue
import re

from supervisely.worker_api.agent_api import AgentAPI
from supervisely.worker_proto import worker_api_pb2 as api_proto
from supervisely.function_wrapper import function_wrapper
from supervisely._utils import take_with_default
from supervisely.sly_logger import logger as default_logger
from supervisely.sly_logger import EventType
from supervisely.app.v1.constants import STATE, CONTEXT, STOP_COMMAND, IMAGE_ANNOTATION_EVENTS
from supervisely.api.api import Api
from supervisely.io.fs import file_exists, mkdir, list_files, get_file_name_with_ext
from supervisely.io.json import load_json_file
from supervisely._utils import _remove_sensitive_information
from supervisely.worker_api.agent_rpc import send_from_memory_generator
from supervisely.io.fs_cache import FileCache


# https://www.roguelynn.com/words/asyncio-we-did-it-wrong/


class ConnectionClosedByServerException(Exception):
    pass


class AppCommandNotFound(Exception):
    pass


REQUEST_ID = 'request_id'
SERVER_ADDRESS = 'SERVER_ADDRESS'
API_TOKEN = 'API_TOKEN'
REQUEST_DATA = "request_data"
AGENT_TOKEN = "AGENT_TOKEN"


def _default_stop(api: Api, task_id, context, state, app_logger):
    app_logger.info('Stop app', extra={'event_type': EventType.APP_FINISHED})


class AppService:
    NETW_CHUNK_SIZE = 1048576
    QUEUE_MAX_SIZE = 2000  # Maximum number of in-flight requests to avoid exhausting server memory.
    DEFAULT_EVENTS = [STOP_COMMAND, *IMAGE_ANNOTATION_EVENTS]

    def __init__(self, logger=None, task_id=None, server_address=None, agent_token=None, ignore_errors=False,
                 ignore_task_id=False):
        self._ignore_task_id = ignore_task_id
        self.logger = take_with_default(logger, default_logger)
        self._ignore_errors = ignore_errors
        self.task_id = take_with_default(task_id, int(os.environ["TASK_ID"]))
        self.server_address = take_with_default(server_address, os.environ[SERVER_ADDRESS])
        self.agent_token = take_with_default(agent_token, os.environ[AGENT_TOKEN])
        self.public_api = Api.from_env(ignore_task_id=self._ignore_task_id)
        self._app_url = self.public_api.app.get_url(self.task_id)
        self._session_dir = "/sessions/{}".format(self.task_id)
        self._template_path = None
        debug_app_dir = os.environ.get("DEBUG_APP_DIR", "")
        if debug_app_dir != "":
            self._session_dir = debug_app_dir
        mkdir(self.data_dir)

        self.cache_dir = os.path.join("/apps_cache")
        debug_cache_dir = os.environ.get("DEBUG_CACHE_DIR", "")
        if debug_cache_dir != "":
            self.cache_dir = debug_cache_dir
        mkdir(self.cache_dir)
        self.cache = FileCache(name="FileCache", storage_root=self.cache_dir)

        self.api = AgentAPI(token=self.agent_token, server_address=self.server_address, ext_logger=self.logger)
        self.api.add_to_metadata('x-task-id', str(self.task_id))

        self.callbacks = {}
        self.periodic_items = {}

        self.processing_queue = queue.Queue()  # (maxsize=self.QUEUE_MAX_SIZE)
        self.logger.debug('App is created', extra={"task_id": self.task_id, "server_address": self.server_address})

        self._ignore_stop_for_debug = False
        self._error = None
        self.stop_event = asyncio.Event()
        self.has_ui = False

    def _run_executors(self):
        self.executor = concurrent.futures.ThreadPoolExecutor()
        self.loop = asyncio.get_event_loop()
        # May want to catch other signals too
        signals = (signal.SIGHUP, signal.SIGTERM, signal.SIGINT, signal.SIGQUIT)
        for s in signals:
            self.loop.add_signal_handler(s, lambda s=s: asyncio.create_task(self._shutdown(signal=s)))
        # comment out the line below to see how unhandled exceptions behave
        self.loop.set_exception_handler(self.handle_exception)

    def handle_exception(self, loop, context):
        # context["message"] will always be there; but context["exception"] may not
        msg = context.get("exception", context["message"])
        if isinstance(msg, Exception):
            # self.logger.error(traceback.format_exc(), exc_info=True, extra={'exc_str': str(msg), 'future_info': context["future"]})
            self.logger.error(msg, exc_info=True, extra={'future_info': context["future"]})
        else:
            self.logger.error("Caught exception: {}".format(msg))

        self.logger.info("Shutting down...")
        asyncio.create_task(self._shutdown())

    @property
    def session_dir(self):
        return self._session_dir

    @property
    def repo_dir(self):
        return os.path.join(self._session_dir, "repo")

    @property
    def data_dir(self):
        return os.path.join(self._session_dir, "data")

    @property
    def app_url(self):
        return self._app_url

    def _add_callback(self, callback_name, func):
        self.callbacks[callback_name] = func

    def callback(self, callback_name):
        """A decorator that is used to register a view function for a
        given application command.  This does the same thing as :meth:`add_callback`
        but is intended for decorator usage::
            @app.callback('calc')
            def calc_func():
                return 'Hello World'
        :param callback_name: the command name as string
        """

        def decorator(f):
            self._add_callback(callback_name, f)

            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                f(*args, **kwargs)

            return wrapper

        return decorator

    def call_periodic_function_sync(self, f, period):
        while True:
            then = time.time()

            try:
                f(api=self.public_api,
                  task_id=self.task_id)
            except Exception as ex:
                tb = traceback.format_exc()
                self.logger.error(f"Exception in periodic function: {f.__name__}\n"
                                  f"{tb}")
                self.logger.info("App will be stopped due to error")

                asyncio.run_coroutine_threadsafe(self._shutdown(error=ex), self.loop)

            elapsed = time.time() - then

            if (period - elapsed) > 0:
                # await asyncio.sleep(period - elapsed)
                time.sleep(period - elapsed)

    # async def call_periodic_function(self, period, f):

    async def scheduler(self):
        self.logger.info("Starting scheduler")

        for f, seconds in self.periodic_items.items():
            asyncio.ensure_future(
                self.loop.run_in_executor(self.executor, self.call_periodic_function_sync, f, seconds), loop=self.loop
            )
            # self.loop.create_task(self.call_periodic_function(seconds, f))

    def _add_periodic(self, seconds, f):
        self.periodic_items[f] = seconds

    def periodic(self, seconds):
        """A decorator that is used to call functions periodically
            @app.periodic(seconds=5)
            def log_message_periodically():
                sly.logger.info('periodically message')
        :param seconds: interval of function call in seconds
        """

        def decorator(f):
            self._add_periodic(seconds, f)

            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                f(*args, **kwargs)

            return wrapper

        return decorator

    def handle_message_sync(self, request_msg):
        try:

            state = request_msg.get(STATE, None)
            context = request_msg.get(CONTEXT, None)
            if context is not None:
                context["request_id"] = request_msg.get("request_id", "")
            command = request_msg["command"]
            user_api_token = request_msg["api_token"]
            user_public_api = Api(self.server_address, user_api_token, retry_count=5, external_logger=self.logger,
                                  ignore_task_id=self._ignore_task_id)
            self.logger.trace("Event", extra={"request_msg": request_msg})

            if command == STOP_COMMAND:
                self.logger.info("APP receives stop signal from user")
                self.stop_event.set()

            if command == STOP_COMMAND and command not in self.callbacks:
                _default_stop(user_public_api, self.task_id, context, state, self.logger)
                if self._ignore_stop_for_debug is False:
                    # self.stop()
                    asyncio.run_coroutine_threadsafe(self._shutdown(), self.loop)
                    return
                else:
                    self.logger.info("STOP event is ignored ...")
            elif command in AppService.DEFAULT_EVENTS and command not in self.callbacks:
                raise AppCommandNotFound("App received default command {!r}. Use decorator \"callback\" to handle it."
                                         .format(command))
            elif command not in self.callbacks:
                raise AppCommandNotFound("App received unhandled command {!r}. Use decorator \"callback\" to handle it."
                                         .format(command))

            if command == STOP_COMMAND:
                if self._ignore_stop_for_debug is False:
                    self.callbacks[command](api=user_public_api,
                                            task_id=self.task_id,
                                            context=context,
                                            state=state,
                                            app_logger=self.logger)
                    asyncio.run_coroutine_threadsafe(self._shutdown(), self.loop)
                    return
                else:
                    self.logger.info("STOP event is ignored ...")
            else:
                self.callbacks[command](api=user_public_api,
                                        task_id=self.task_id,
                                        context=context,
                                        state=state,
                                        app_logger=self.logger)
        except AppCommandNotFound as e:
            self.logger.debug(repr(e), exc_info=False)
        except Exception as e:
            if self._ignore_errors is False:
                self.logger.error(traceback.format_exc(), exc_info=True, extra={
                    'main_name': command,
                    'exc_str': repr(e),
                    'event_type': EventType.TASK_CRASHED
                })
                self.logger.info("App will be stopped due to error")
                # asyncio.create_task(self._shutdown(error=e))
                asyncio.run_coroutine_threadsafe(self._shutdown(error=e), self.loop)
            else:
                self.logger.error(traceback.format_exc(), exc_info=True, extra={'exc_str': repr(e)})
                if self.has_ui:
                    self.show_modal_window("Oops! Something went wrong, please try again or contact tech support."
                                           " Find more info in the app logs.", level="error")

    def consume_sync(self):
        while True:
            request_msg = self.processing_queue.get()
            to_log = _remove_sensitive_information(request_msg)
            self.logger.debug('FULL_TASK_MESSAGE', extra={'task_msg': to_log})
            # asyncio.run_coroutine_threadsafe(self.handle_message(request_msg), self.loop)
            asyncio.ensure_future(
                self.loop.run_in_executor(self.executor, self.handle_message_sync, request_msg), loop=self.loop
            )

    async def consume(self):
        self.logger.info("Starting consumer")
        asyncio.ensure_future(
            self.loop.run_in_executor(self.executor, self.consume_sync), loop=self.loop
        )

    def publish_sync(self, initial_events=None):
        if initial_events is not None:
            for event_obj in initial_events:
                event_obj["api_token"] = os.environ[API_TOKEN]
                self.processing_queue.put(event_obj)

        for gen_event in self.api.get_endless_stream('GetGeneralEventsStream', api_proto.GeneralEvent,
                                                     api_proto.Empty()):
            try:
                data = {}
                if gen_event.data is not None and gen_event.data != b'':
                    data = json.loads(gen_event.data.decode('utf-8'))

                event_obj = {REQUEST_ID: gen_event.request_id, **data}
                self.processing_queue.put(event_obj)
            except Exception as error:
                self.logger.warning('App exception: ', extra={"error_message": repr(error)})

        raise ConnectionClosedByServerException('Requests stream to a deployed model closed by the server.')

    async def publish(self, initial_events=None):
        self.logger.info("Starting publisher")
        asyncio.ensure_future(
            self.loop.run_in_executor(self.executor, self.publish_sync, initial_events), loop=self.loop
        )

    def run(self, template_path=None, data=None, state=None, initial_events=None):
        if template_path is None:
            template_path = self.get_template_path()
            self.logger.info(f"App template path: {template_path}")
        else:
            self._template_path = template_path

        if template_path is None:
            template_path = os.path.join(self.repo_dir, "src/gui.html")

        if not file_exists(template_path):
            self.logger.info("App will be running without GUI", extra={"app_url": self.app_url})
            template = ""
        else:
            with open(template_path, 'r') as file:
                template = file.read()
            self.has_ui = True

        self.public_api.app.initialize(self.task_id, template, data, state)
        self.logger.info("Application session is initialized", extra={"app_url": self.app_url})

        try:
            self._run_executors()
            self.loop.create_task(self.publish(initial_events), name="Publisher")
            self.loop.create_task(self.consume(), name="Consumer")
            self.loop.create_task(self.scheduler(), name="Scheduler")
            self.loop.run_forever()
        finally:
            self.loop.close()
            self.logger.info("Successfully shutdown the APP service.")

        if self._error is not None:
            raise self._error

    def stop(self, wait=True):
        # @TODO: add timeout
        if wait is True:
            event_obj = {"command": "stop", "api_token": os.environ[API_TOKEN]}
            self.processing_queue.put(event_obj)
        else:
            self.logger.info('Stop app (force, no wait)', extra={'event_type': EventType.APP_FINISHED})
            # asyncio.create_task(self._shutdown())
            asyncio.run_coroutine_threadsafe(self._shutdown(), self.loop)

    async def _shutdown(self, signal=None, error=None):
        """Cleanup tasks tied to the service's shutdown."""
        if signal:
            self.logger.info(f"Received exit signal {signal.name}...")
        self.logger.info("Nacking outstanding messages")
        tasks = [t for t in asyncio.all_tasks() if t is not
                 asyncio.current_task()]

        [task.cancel() for task in tasks]

        self.logger.info(f"Cancelling {len(tasks)} outstanding tasks")
        await asyncio.gather(*tasks, return_exceptions=True)

        self.logger.info("Shutting down ThreadPoolExecutor")
        self.executor.shutdown(wait=False)

        self.logger.info(f"Releasing {len(self.executor._threads)} threads from executor")
        for thread in self.executor._threads:
            try:
                thread._tstate_lock.release()
            except Exception:
                pass

        self.logger.info(f"Flushing metrics")
        self.loop.stop()

        if error is not None:
            self._error = error

    def send_response(self, request_id, data):
        out_bytes = json.dumps(data).encode('utf-8')
        self.api.put_stream_with_data('SendGeneralEventData',
                                      api_proto.Empty,
                                      send_from_memory_generator(out_bytes, 1048576),
                                      addit_headers={'x-request-id': request_id})

    def show_modal_window(self, message, level="info"):
        all_levels = ["warning", "info", "error"]
        if level not in all_levels:
            raise ValueError("Unknown level {!r}. Supported levels: {}".format(level, all_levels))

        if level == "info":
            self.logger.info(message)
        elif level == "warning":
            self.logger.warning(message)
        elif level == "error":
            self.logger.error(message, exc_info=True)

        self.public_api.app.set_field(self.task_id, "data.notifyDialog", {
            "type": level,
            "message": message
        })

    def get_template_path(self):
        if self._template_path is not None:
            return self._template_path
        config_path = os.path.join(self.repo_dir, os.environ.get("CONFIG_DIR", ""), 'config.json')
        if file_exists(config_path):
            config = load_json_file(config_path)
            self._template_path = config.get('gui_template', None)
            if self._template_path is None:
                self.logger.info("there is no gui_template field in config.json")
            else:
                self._template_path = os.path.join(self.repo_dir, self._template_path)
                if not file_exists(self._template_path):
                    self._template_path = os.path.join(os.path.dirname(sys.argv[0]), 'gui.html')
        if self._template_path is None:
            self._template_path = os.path.join(os.path.dirname(sys.argv[0]), 'gui.html')
        if file_exists(self._template_path):
            return self._template_path
        return None

    def compile_template(self, root_source_dir=None):
        if root_source_dir is None:
            root_source_dir = self.repo_dir

        def _my_replace_function(match):
            to_replace = match.group(0)
            part_path = match.group(1)
            with open(os.path.join(root_source_dir, part_path), 'r') as file:
                part = file.read()
            return part

        template_path = self.get_template_path()
        if template_path is None:
            self.logger.warning("HTML Template for compilation not found")
            return
        with open(template_path, 'r') as file:
            template = file.read()

        regex = r"{\%.*include.*'(.*)'.*\%}"
        result = re.sub(regex, lambda m: _my_replace_function(m), template, 0, re.MULTILINE)
        res_path = os.path.join(self.data_dir, "gui.html")
        with open(os.path.join(self.data_dir, "gui.html"), "w") as file:
            file.write(result)
        self._template_path = res_path
        self.logger.info(f"Compiled template is saved to {res_path}")
        return res_path

    def ignore_errors_and_show_dialog_window(self):
        def decorator(f):
            @functools.wraps(f)
            def wrapper(*args, **kwargs):
                try:
                    f(*args, **kwargs)
                except Exception as e:
                    self.logger.error(
                        f"please, contact support: task_id={self.task_id}, {repr(e)}",
                        exc_info=True,
                        extra={
                            'exc_str': str(e),
                        }
                    )
                    self.show_modal_window(f"Oops! Something went wrong, please try again or contact tech support."
                                           f" Find more info in the app logs. Error: {repr(e)}", level="error")

            return wrapper

        return decorator



# coding: utf-8

from __future__ import annotations
import skvideo.io
from typing import Tuple, List, Dict, Optional

import os
from supervisely import logger as default_logger
from supervisely.io.fs import get_file_name, get_file_ext
from supervisely._utils import rand_str


# Do NOT use directly for video extension validation. Use is_valid_ext() /  has_valid_ext() below instead.
ALLOWED_VIDEO_EXTENSIONS = ['.avi', '.mp4', '.3gp', '.flv', '.webm', '.wmv', '.mov', '.mkv']


_SUPPORTED_CONTAINERS = {'mp4', 'webm', 'ogg', 'ogv'}
_SUPPORTED_CODECS = {'h264', 'vp8', 'vp9'}


class VideoExtensionError(Exception):
    pass


class UnsupportedVideoFormat(Exception):
    pass


class VideoReadException(Exception):
    pass


def is_valid_ext(ext: str) -> bool:
    """
    Checks if given extension is supported.

    :param ext: Video file extension.
    :type ext: str
    :return: bool
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        sly.video.is_valid_ext(".mp4")  # True
        sly.video.is_valid_ext(".jpeg") # False
    """
    return ext.lower() in ALLOWED_VIDEO_EXTENSIONS


def has_valid_ext(path: str) -> bool:
    """
    Checks if Video file from given path has supported extension.

    :param path: Path to Video file.
    :type path: str
    :return: bool
    :rtype: :class:`bool`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        video_path = "/home/admin/work/videos/Cars/ds0/video/6x.mp4"
        sly.video.has_valid_ext(video_path) # True
    """
    return is_valid_ext(os.path.splitext(path)[1])


def validate_ext(ext: str):
    """
    Raises error if given extension is not supported.

    :param ext: Extension.
    :type ext: str
    :raises: :class:`UnsupportedVideoFormat` if given extension is not supported.
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

       import supervisely as sly

        sly.video.validate_ext(".jpeg")
        # Unsupported video extension: .jpeg.
        # Only the following extensions are supported: ['.avi', '.mp4', '.3gp', '.flv', '.webm', '.wmv', '.mov', '.mkv'].
    """
    if not is_valid_ext(ext):
        raise UnsupportedVideoFormat('Unsupported video extension: {}. Only the following extensions are supported: {}.'
                                     .format(ext, ALLOWED_VIDEO_EXTENSIONS))


def get_image_size_and_frames_count(path: str) -> Tuple[Tuple[int, int], int]:
    """
    Gets image size and number of frames from Video file.

    :param path: Path to Video file.
    :type path: str
    :return: Image size and number of Video frames.
    :rtype: :class:`Tuple[Tuple[int, int], int]`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        video_path = "/home/admin/work/videos/Cars/ds0/video/6x.mp4"
        video_info = sly.video.get_image_size_and_frames_count(video_path)
        print(video_info)
        # Output: ((720, 1280), 152)
    """
    vreader = skvideo.io.FFmpegReader(path)
    vlength = vreader.getShape()[0]
    img_height = vreader.getShape()[1]
    img_width = vreader.getShape()[2]

    img_size = (img_height, img_width)

    return img_size, vlength


def validate_format(path: str) -> None:
    """
    Raise error if Video file from given path couldn't be read or file extension is not supported.

    :param path: Path to Video file.
    :type path: str
    :raises: :class:`VideoReadException` if Video file from given path couldn't be read or file extension is not supported
    :return: None
    :rtype: :class:`NoneType`
    :Usage example:

     .. code-block:: python

        import supervisely as sly

        video_path = "/home/paul/work/sphinx-docs/supervisely_py/docs/source/debug/video/Prius_360/ds0/video/video.jpg"
        sly.video.validate_format(video_path)
        # Unsupported video extension: .jpg. Only the following extensions are supported: ['.avi', '.mp4', '.3gp', '.flv', '.webm', '.wmv', '.mov', '.mkv'].
    """
    try:
        get_image_size_and_frames_count(path)
    except Exception as e:
        raise VideoReadException(
            'Error has occured trying to read video {!r}. Original exception message: {!r}'.format(path, str(e)))

    validate_ext(os.path.splitext(path)[1])


def _check_video_requires_processing(video_info, stream_info):
    """
    Check if video need container or codec processing
    :param video_info: dict
    :param stream_info: dict
    :return: bool
    """
    need_process_container = True
    for name in video_info["meta"]["formatName"].split(','):
        name = name.strip().split('.')[-1]
        if name in _SUPPORTED_CONTAINERS:
            need_process_container = False
            break

    need_process_codec = True
    codec = stream_info["codecName"]
    if codec in _SUPPORTED_CODECS:
        need_process_codec = False

    if (need_process_container is False) and (need_process_codec is False):
        return False

    return True


def count_video_streams(all_streams: List[Dict]) -> int:
    """
    Count number of video streams in video.

    :param all_streams: List of Video file audio and video streams.
    :type all_streams: List[dict]
    :return: Number of video streams in Video file
    :rtype: :class:`int`
    """
    count = 0
    for stream_info in all_streams:
        if stream_info["codecType"] == "video":
            count += 1
    return count


def get_video_streams(all_streams: List[Dict]) -> List:
    """
    Get list of video streams from given list of all streams.

    :param all_streams: List of Video file audio and video streams.
    :type all_streams: List[dict]
    :return: List of video streams in Video file.
    :rtype: :class:`list`
    """
    video_streams = []
    for stream_info in all_streams:
        if stream_info["codecType"] == "video":
            video_streams.append(stream_info)
    return video_streams


def warn_video_requires_processing(file_name: str, logger: Optional[default_logger]=None) -> None:
    """
    Create logger if it was not there and displays message about the need for transcoding.

    :param file_name: Video file name.
    :type file_name: str
    :param logger: Logger object.
    :type logger: logger
    :return: None
    :rtype: :class:`NoneType`
    """
    if logger is None:
        logger = default_logger
    logger.warning("Video Stream {!r} is skipped: requires transcoding. Transcoding is supported only in Enterprise Edition (EE)".format(file_name))


def gen_video_stream_name(file_name: str, stream_index: int) -> str:
    """
    Create name to video stream from given filename and index of stream.

    :param file_name: Video file name.
    :type file_name: str
    :param stream_index: Stream index.
    :type stream_index: int
    :return: str
    :rtype: str
    :Usage example:

     .. code-block:: python

        stream_name = gen_video_stream_name('my_video.mp4', 2)
        print(stream_name)
        # Output: my_video_stream_2_CULxO.mp4
    """
    return "{}_stream_{}_{}{}".format(get_file_name(file_name), stream_index, rand_str(5), get_file_ext(file_name))


def get_info(video_path: str, cpu_count: Optional[int]=None) -> Dict:
    """
    Get information about video from given path.

    :param video_path: Video file path.
    :type video_path: str
    :param cpu_count: CPU count.
    :type cpu_count: int
    :raises: :class:`ValueError` if no video streams found.
    :return: Information about video
    :rtype: :class:`Dict`
    :Usage example:

     .. code-block:: python

        from supervisely.video.video import get_info
        video_info = get_info('/home/video/1.mp4')
        print(json.dumps(video_info, indent=4))
        # Output: {
        #     "streams": [
        #         {
        #             "index": 0,
        #             "width": 1920,
        #             "height": 1080,
        #             "duration": 16.666667,
        #             "rotation": 0,
        #             "codecName": "mpeg4",
        #             "codecType": "video",
        #             "startTime": 0,
        #             "framesCount": 500,
        #             "framesToTimecodes": [
        #                 0.0,
        #                 0.033333,
        #                 0.066667,
        #                 0.1,
        #                   ...
        #                 16.566667,
        #                 16.6,
        #                 16.633333
        #             ]
        #         }
        #     ],
        #     "formatName": "mov,mp4,m4a,3gp,3g2,mj2",
        #     "duration": 16.667,
        #     "size": "61572600"
        # }
    """
    import pathlib
    import subprocess, os, ast, math
    from subprocess import PIPE

    def rotate_dimensions(width, height, rotation):
        cur_angle = rotation * math.pi / 180
        c = math.cos(cur_angle)
        s = math.sin(cur_angle)
        w = round(abs(width * c - height * s))
        h = round(abs(width * s + height * c))
        return w, h

    if cpu_count is None:
        cpu_count = os.cpu_count()

    current_dir = pathlib.Path(__file__).parent.absolute()
    session = subprocess.Popen(['sh', os.path.join(current_dir, 'get_video_info.sh'), video_path, str(cpu_count)], stdout=PIPE, stderr=PIPE)
    stdout, stderr = session.communicate()
    if len(stderr) != 0:
        raise RuntimeError(stderr.decode("utf-8"))
    video_meta = ast.literal_eval(stdout.decode('utf-8'))

    frames_to_timecodes = []
    has_video = False
    audio_stream_info = None

    for frame in video_meta['frames']:
        if frame['stream_index'] == 0:
            frames_to_timecodes.append(float(frame.get('pkt_pts_time', frame.get('pts_time'))))

    stream_infos = []
    for stream in video_meta['streams']:
        if stream['codec_type'] == 'video':
            has_video = True
            stream_info = {'index': stream['index'], 'width': stream['width'], 'height': stream['height'],
                           'duration': float(stream['duration']), 'rotation': 0, 'codecName': stream['codec_name'],
                           'codecType': stream['codec_type'], 'startTime': int(float(stream['start_time'])),
                           'framesCount': len(frames_to_timecodes), 'framesToTimecodes': frames_to_timecodes}
            side_data_list = stream.get('side_data_list', None)
            if side_data_list:
                for data in side_data_list:
                    rotation = data.get('rotation', None)
                if rotation:
                    stream_info['rotation'] = rotation
                    width, height = rotate_dimensions(stream_info['width'], stream_info['height'], rotation)
                    stream_info['originalWidth'] = stream_info['width']
                    stream_info['originalHeight'] = stream_info['height']
                    stream_info['width'] = width
                    stream_info['height'] = height
        elif stream['codec_type'] == 'audio':
            stream_info = {'index': stream['index'], 'channels': stream['channels'],
                           'duration': float(stream['duration']), 'codecName': stream['codec_name'],
                           'codecType': stream['codec_type'], 'startTime': int(float(stream['start_time'])),
                           'sampleRate': int(stream['sample_rate'])}
        else:
            continue
        stream_infos.append(stream_info)

    if has_video is False:
        raise ValueError('No video streams found')

    result = {'streams': stream_infos,
              'formatName': video_meta['format']['format_name'],
              'duration': float(video_meta['format']['duration']),
              'size': video_meta['format']['size']}

    return result



# from .compare_gallery import CompareGallery
# from .chart import Chart
# from .empty_grid_gallery import EmptyGridGallery
# from .predictions_dynamics_gallery import PredictionsDynamicsGallery
# from .progress_bar import ProgressBar
# from .single_image_gallery import SingleImageGallery



class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        local = kwargs.pop("__local__", False)
        if local is False:
            if cls not in cls._instances:
                cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            return cls._instances[cls]
        else:
            return super(Singleton, cls).__call__(*args, **kwargs)



import sys

_import_failed = False
try:
    from supervisely.app.content import StateJson, DataJson
    from supervisely.app.content import get_data_dir
    import supervisely.app.fastapi as fastapi
    import supervisely.app.widgets as widgets
except (ImportError, ModuleNotFoundError) as e:
    _import_failed = True
    pass


def __getattr__(name):
    if _import_failed is True:
        raise ModuleNotFoundError(
            'No module named supervisely.app, please install dependencies with "pip install supervisely[apps]"'
        )
    return getattr(sys.modules[__name__], name)
